import logging

import numpy as np
from isimip_utils.pandas import compute_average, create_label, group_by_day, group_by_month, normalize
from isimip_utils.parameters import copy_placeholders, get_placeholders, join_parameters
from isimip_utils.plot import format_title, plot_grid, plot_line, plot_map, save_index, save_plot
from isimip_utils.xarray import open_dataset, to_dataframe

from .config import settings
from .models import Dataset, Extraction, Figure

logger = logging.getLogger(__name__)


def create_plots(periods, regions, aggregations, plots):
    logger.info('Creating plots')

    index_paths = set()

    for period in periods:
        for region in regions:
            for aggregation in aggregations:
                for plot in plots:
                    for path in settings.PATHS:
                        for figs_permutation in settings.FIGS_PERMUTATIONS:
                            figure_placeholders = copy_placeholders(
                                get_placeholders(settings.FIGS_PARAMETERS, figs_permutation),
                                join_parameters(settings.GRID_PARAMETERS, max_count=4),
                                join_parameters(settings.PLOT_PARAMETERS, max_count=4)
                            )

                            figure = Figure(path, figure_placeholders, period, region, aggregation, plot)

                            if settings.FORCE or not figure.exists():
                                charts = {}
                                for grid_permutation in settings.GRID_PERMUTATIONS:
                                    for plot_permutation in settings.PLOT_PERMUTATIONS:
                                        dataset_placeholders = copy_placeholders(
                                            get_placeholders(settings.FIGS_PARAMETERS, figs_permutation),
                                            get_placeholders(settings.GRID_PARAMETERS, grid_permutation),
                                            get_placeholders(settings.PLOT_PARAMETERS, plot_permutation),
                                        )
                                        dataset = Dataset(path, dataset_placeholders)

                                        extraction = Extraction(dataset, period, region, aggregation)
                                        if extraction.exists():
                                            with open_dataset(extraction.full_path) as ds:
                                                df = get_dataframe(ds, plot, plot_permutation)
                                                if df is not None:
                                                    chart = get_chart(df, plot, labels=plot_permutation)
                                                    charts[grid_permutation + plot_permutation] = chart

                                if charts:
                                    empty_chart = get_chart(df, plot, empty=True)

                                    chart = plot_grid(
                                        settings.GRID_PERMUTATIONS, settings.PLOT_PERMUTATIONS,
                                        charts, empty_chart, **settings.PLOT_RESOLVE_SCALE
                                    ).properties(title=get_title(figs_permutation, period, region, aggregation, plot))

                                    save_plot(chart, figure.full_path)

                                    index_paths.add(figure.full_path.parent)

    if settings.PLOT_INDEX:
        for parent_path in index_paths:
            save_index(parent_path / 'index.html')


def get_dataframe(ds, plot, labels):
    try:
        df = to_dataframe(ds)
    except ValueError as e:
        logger.error(f'error converting dataset: "{e}"')
        return

    columns = set(df.attrs['coords'].keys())

    if plot.type == 'value':
        if columns.intersection(plot.columns):
            df = df.replace(0, np.nan)
            df = create_label(df, labels)
            return df
        else:
            logger.warn(f'incorrect columns {columns} for plot "{plot.specifier}"')

    elif plot.type == 'annual':
        if columns.intersection(plot.columns):
            df = compute_average(df, area=False)
            df = create_label(df, labels)
            return df
        else:
            logger.warn(f'incorrect columns {columns} for plot "{plot.specifier}"')

    elif plot.type == 'dayofyear':
        if columns.intersection(plot.columns):
            df = group_by_day(df)
            df = normalize(df)
            df = create_label(df, labels)
            return df
        else:
            logger.warn(f'incorrect columns {columns} for plot "{plot.specifier}"')

    elif plot.type == 'monthofyear':
        if columns.intersection(plot.columns):
            df = group_by_month(df)
            df = normalize(df)
            df = create_label(df, labels)
            return df
        else:
            logger.warn(f'incorrect columns {columns} for plot "{plot.specifier}"')

    elif plot.type == 'map':
        if columns.intersection(plot.columns):
            df = create_label(df, labels)
            return df
        else:
            logger.warn(f'incorrect columns {columns} for plot "{plot.specifier}"')

    else:
        logger.error(f'unknown type "{plot.type}" for plot "{plot.specifier}"')


def get_chart(df, plot, labels=None, **kwargs):
    kwargs = dict(legend=bool(labels), color_scheme=settings.COLOR_SCHEME, **kwargs)

    if settings.PRIMARY and labels:
        if set(settings.PRIMARY).intersection(set(labels)):
            kwargs.update(strokeWidth=3)
        else:
            kwargs.update(strokeWidth=0.5)

    if plot.type == 'value':
        return plot_line(df, y_format='.1e', **kwargs)

    elif plot.type == 'annual':
        return plot_line(df, y_format='.1e', interpolate='step-after', **kwargs)

    elif plot.type == 'dayofyear':
        return plot_line(df, **kwargs)

    elif plot.type == 'monthofyear':
        return plot_line(df, **kwargs)

    elif plot.type == 'map':
        return plot_map(df, color_format='.1e', **kwargs)


def get_title(permutation, period, region, aggregation, plot):
    args = list(permutation)

    if period.type != 'auto':
        args.append(period.specifier)

    if region.type != 'global':
        args.append(region.specifier)

    if aggregation.type != 'value':
        args.append(aggregation.specifier)

    if plot.type != 'value':
        args.append(plot.specifier)

    return format_title(args)



from isimip_utils.cli import ArgumentParser, parse_list, parse_locations, parse_path, setup_logs

from . import VERSION
from .cli import ArgumentAction
from .config import settings
from .extractions import create_extractions, fetch_extractions
from .models import Aggregation, Period, Plot, Region
from .plots import create_plots


def main():
    config_parser = ArgumentParser(add_help=False)

    config_parser.add_argument('-c', '--config', dest='config_path', type=parse_path,
                        help='Path to an additional config file, updating default and CLI arguments and options.')

    config_args, remaining_args = config_parser.parse_known_args()

    parser = ArgumentParser(prog='isimip-ea', parents=[config_parser])

    parser.add_argument('paths', nargs='*', action=ArgumentAction,
                        help='Paths of the datasets to process, can contain placeholders, e.g. {model}')
    parser.add_argument('parameters', nargs='*', action=ArgumentAction,
                        help='Values for the placeholders in the from key=value1,value2,...')

    parser.add_argument('--datasets-path', dest='datasets_path', type=parse_path, default='.',
                        help='Base path for the input datasets')
    parser.add_argument('--extractions-path', dest='extractions_path', type=parse_path, default='.',
                        help='Base path for the created extractions')
    parser.add_argument('--plots-path', dest='plots_path', type=parse_path, default='.',
                        help='Base path for the created plots')

    parser.add_argument('-d', '--dates', dest='dates', type=parse_list, default='auto',
                        help='Extract only specific dates or periods (comma separated, format: YYYY, YYYYMMDD, '
                             'YYYY-YYYY, YYYYMMDD-YYYYMMDD)')
    parser.add_argument('-r', '--regions', dest='regions', type=parse_list, default='global',
                        help='Extract only specific regions (comma separated, automatically selected '
                             'from --regions-locations)')
    parser.add_argument('-a', '--aggregations', dest='aggregations', type=parse_list, default='mean',
                        help='Perform aggregations when extracting (comma separated: value, mean, std, sum, min, max, '
                             'count, meanmap, countmap)')
    parser.add_argument('-p', '--plots', dest='plots', type=parse_list, default='value',
                        help='Select specific plots (comma separated: value, annual, dayofyear, monthofyear, map)')

    parser.add_argument('-f', '--force', dest='force', action='store_true', default=False,
                        help='Overwrite existing files')
    parser.add_argument('-l', '--load', dest='load', action='store_true', default=False,
                        help='Load NetCDF datasets in memory, useful for point extractions')

    parser.add_argument('--fetch-only', dest='fetch_only', action='store_true', default=False,
                        help='Only fetch extractions')
    parser.add_argument('--extractions-only', dest='extractions_only', action='store_true', default=False,
                        help='Only create extractions')
    parser.add_argument('--plots-only', dest='plots_only', action='store_true', default=False,
                        help='Only create plots')

    parser.add_argument('--gridarea', dest='gridarea', type=parse_path,
                        help='Use a CDO gridarea file instead of computing the gridarea when computing means')

    parser.add_argument('--plot-format', dest='plot_format', default='svg',
                        help='File format for plots [default: svg].')
    parser.add_argument('--plot-index', dest='plot_index', default=True,
                        help='Create an index.html file when creating plots.')
    parser.add_argument('--primary', dest='primary',
                        help='Treat these placeholders as primary and plot them in color [default: all]')
    parser.add_argument('--grid-placeholders', dest='grid_placeholders', type=parse_list, default="",
                        help='Parameters which are used as dimensions of the plot grid.')
    parser.add_argument('--figure-placeholders', dest='figure_placeholders', type=parse_list, default="",
                        help='Parameters for which separate figures are created')
    parser.add_argument('--figure-path', dest='figure_path', type=parse_path, default=None,
                        help='Custom paths for the created figures, can contain placeholders')
    parser.add_argument('--color-scheme', dest='color_scheme', default='category20',
                        help='Color scheme to use for plots [default: category20].')

    parser.add_argument('--independent-x', dest='independent_x', action='store_true', default=False,
                        help='Use independent x axis in plots')
    parser.add_argument('--independent-y', dest='independent_y', action='store_true', default=False,
                        help='Use independent y axis in plots')
    parser.add_argument('--shared-color', dest='shared_color', action='store_true', default=False,
                        help='Use shared color scale in plots')

    parser.add_argument('--protocol-location', dest='protocol_locations', type=parse_locations,
                        default='https://protocol.isimip.org https://protocol2.isimip.org',
                        help='URL or file path to the protocol')
    parser.add_argument('--regions-location', dest='regions_locations', type=parse_locations,
                        help='Use the provided files to create the regions.')
    parser.add_argument('--extractions-locations', dest='extractions_locations', type=parse_locations,
                        default='https://files.isimip.org/extractions/',
                        help='URL or file path to the locations of extractions to fetch')

    parser.add_argument('--log-level', dest='log_level', default='WARN',
                        help='Log level (ERROR, WARN, INFO, or DEBUG)')
    parser.add_argument('--log-file', dest='log_file',
                        help='Path to the log file')
    parser.add_argument('--show-time', dest='show_time', action='store_true', default=False,
                        help='show time in console logs')
    parser.add_argument('--show-path', dest='show_path', action='store_true', default=False,
                        help='show path in console logs')

    parser.add_argument('-V', '--version', action='version', version=VERSION)

    args = parser.parse_args(remaining_args, config_path=config_args.config_path)

    setup_logs(log_level=args.log_level, log_file=args.log_file, show_time=args.show_time, show_path=args.show_path)

    settings.from_dict(vars(args))

    if not settings.PATHS:
        parser.error('You need to provide at least one path.')

    periods = [Period(value) for value in settings.DATES]
    regions = [Region(specifier) for specifier in settings.REGIONS]
    aggregations = [Aggregation(value) for value in settings.AGGREGATIONS]
    plots = [Plot(value) for value in settings.PLOTS]

    # fetch extractions
    if not settings.EXTRACTIONS_ONLY and not settings.PLOTS_ONLY and settings.EXTRACTIONS_LOCATIONS:
        fetch_extractions(periods, regions, aggregations)

    # create the extractions
    if not settings.FETCH_ONLY and not settings.PLOTS_ONLY:
        create_extractions(periods, regions, aggregations)

    # create the plots
    if not settings.FETCH_ONLY and not settings.EXTRACTIONS_ONLY:
        create_plots(periods, regions, aggregations, plots)



import argparse
import re

from isimip_utils.cli import parse_parameters, parse_path

parameter_pattern = re.compile(r'^.*?=.*?$')


class ArgumentAction(argparse.Action):

    def __call__(self, parser, args, values, option_string=None):
        for value in values:
            match = parameter_pattern.match(value)
            if match:
                try:
                    args.parameters.update(parse_parameters(value))
                except AttributeError:
                    args.parameters = parse_parameters(value)
            else:
                try:
                    args.paths.append(parse_path(value))
                except AttributeError:
                    args.paths = [parse_path(value)]



from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _version

try:
    VERSION = __version__ = _version(__package__)
except PackageNotFoundError:
    VERSION = __version__ = "0.0.0+unknown"



from isimip_utils.config import Settings as BaseSettings
from isimip_utils.parameters import get_permutations
from isimip_utils.utils import cached_property
from isimip_utils.xarray import open_dataset


class Settings(BaseSettings):

    @cached_property
    def WEIGHTS(self):
        if self.GRIDAREA:
            ds = open_dataset(self.GRIDAREA, load=self.LOAD)
            ds = ds.isel(lon=0)
            return ds.cell_area

    @cached_property
    def FIGS_PARAMETERS(self):
        return {key: self.PARAMETERS[key] for key in self.FIGURE_PLACEHOLDERS} if self.PARAMETERS else {}

    @cached_property
    def GRID_PARAMETERS(self):
        return {key: self.PARAMETERS[key] for key in self.GRID_PLACEHOLDERS} if self.PARAMETERS else {}

    @cached_property
    def PLOT_PARAMETERS(self):
        return {
            key: values for key, values in self.PARAMETERS.items()
            if key not in self.FIGURE_PLACEHOLDERS + self.GRID_PLACEHOLDERS
        } if self.PARAMETERS else {}

    @cached_property
    def FIGS_PERMUTATIONS(self):
        return get_permutations(self.FIGS_PARAMETERS) if self.PARAMETERS else [()]

    @cached_property
    def GRID_PERMUTATIONS(self):
        return get_permutations(self.GRID_PARAMETERS) if self.PARAMETERS else [()]

    @cached_property
    def PLOT_PERMUTATIONS(self):
        return get_permutations(self.PLOT_PARAMETERS) if self.PARAMETERS else [()]

    @cached_property
    def PLOT_RESOLVE_SCALE(self):
        return {
            'x': 'independent' if self.INDEPENDENT_X else 'shared',
            'y': 'independent' if self.INDEPENDENT_Y else 'shared',
            'color': 'shared' if self.SHARED_COLOR else 'independent',
        }


settings = Settings()



import logging

from isimip_utils.extractions import (
    compute_aggregation,
    concat_extraction,
    count_values,
    mask_bbox,
    mask_mask,
    select_period,
    select_point,
    select_time,
)
from isimip_utils.xarray import create_mask, open_dataset, write_dataset

from .config import settings
from .models import Dataset, Extraction

logger = logging.getLogger(__name__)


def fetch_extractions(periods, regions, aggregations):
    logger.info('Fetching extractions')

    if not settings.FORCE:
        for dataset in Dataset.all():
            for extraction in Extraction.gather(dataset, periods, regions, aggregations):
                if not extraction.exists():
                    extraction.fetch()


def create_extractions(periods, regions, aggregations):
    logger.info('Creating extractions')

    for dataset in Dataset.all():
        # init extractions dict for this dataset
        extractions = {}

        # continue only if extractions are missing for this dataset
        if settings.FORCE or not all(
            extraction.exists() for extraction in Extraction.gather(dataset, periods, regions, aggregations)
        ):
            for file in dataset.files:
                with open_dataset(file.path, decode_cf=False, load=settings.LOAD) as ds_file:
                    # loop over periods
                    for period in periods:
                        ds_period = extract_period(ds_file, period)

                        # continue only if at least one time step is selected (ds_period is not empty)
                        if ds_period:
                            # loop over regions
                            for region in regions:
                                ds_region = extract_region(ds_period, region)

                                # point extraction does not allow for additional aggregations
                                if region.type == 'point':
                                    extraction = Extraction(dataset, period, region, 'value')
                                    extractions[extraction.path] = concat_extraction(
                                        extractions.get(extraction.path), ds_region
                                    )
                                    continue

                                # continue only if ds_region is not empty
                                if ds_region:
                                    # loop over aggregations
                                    for aggregation in aggregations:
                                        ds_aggregation = extract_aggregation(ds_region, aggregation)

                                        # concat extraction only if ds_aggregation is not empty
                                        if ds_aggregation:
                                            extraction = Extraction(dataset, period, region, aggregation)
                                            extractions[extraction.path] = concat_extraction(
                                                extractions.get(extraction.path), ds_aggregation
                                            )

        # write extractions
        for extraction_path, extraction_ds in extractions.items():
            write_dataset(extraction_ds, settings.EXTRACTIONS_PATH / extraction_path)


def extract_period(ds, period):
    if period.type == 'auto':
        return ds

    elif period.type == 'period':
        return select_period(ds, period.start_time, period.end_time)

    elif period.type == 'date':
        return select_time(ds, period.time)

    else:
        logger.error(f'unknown type "{period.type}" for period "{period.specifier}"')


def extract_region(ds, region):
    if region.type == 'global':
        return ds

    elif region.type == 'bbox':
        return mask_bbox(ds, region.west, region.east, region.south, region.north)

    elif region.type == 'mask':
        return mask_mask(ds, region.mask_ds, region.mask_var)

    elif region.type == 'shape':
        mask = create_mask(ds, region.df, region.layer)
        return mask_mask(ds, mask)

    elif region.type == 'point':
        return select_point(ds, region.lat, region.lon)

    else:
        logger.error(f'unknown type "{region.type}" for region "{region.specifier}"')


def extract_aggregation(ds, aggregation):
    if aggregation.type == 'value':
        return ds

    elif aggregation.type in ('mean', 'std', 'sum', 'min', 'max'):
        return compute_aggregation(ds, aggregation.type, weights=settings.WEIGHTS)

    elif aggregation.type == 'count':
        return count_values(ds)

    elif aggregation.type == 'meanmap':
        return compute_aggregation(ds, 'mean', dim=('time', ))

    elif aggregation.type == 'countmap':
        return count_values(ds, dim=('time', ))

    else:
        logger.error(f'unknown type "{aggregation.type}" for aggregation "{aggregation.specifier}"')



import logging
import re
from datetime import datetime

import pandas as pd
from isimip_utils.xarray import open_dataset

from .config import settings

logger = logging.getLogger(__name__)


def init_period(option):
    if re.match(r'\d{4}-\d{4}', option):
        left, right = option.split('-', 1)
        start_time, end_time = parse_date(left), parse_date(right, start=False)

        return {
            'type': 'period',
            'specifier': f'{start_time}_{end_time}',
            'start_time': start_time,
            'end_time': end_time
        }
    elif re.match(r'\d{4}', option):
        return {
            'type': 'date',
            'specifier': option,
            'time': parse_date(option)
        }

    # if region could not be determined, log error and return
    logger.error(f'could not determine type for period "{option}"')
    return {
        'type': 'unknown',
        'specifier': option
    }


def init_region(value):
    if settings.REGIONS_LOCATIONS:
        for location in settings.REGIONS_LOCATIONS:
            if not location.exists():
                raise RuntimeError(f'{location} does not exist.')

            if location.suffix in ['.json', '.csv']:
                if location.suffix == '.json':
                    df = pd.read_json(location)
                else:
                    df = pd.read_csv(location)

                row = find_row(df, value)
                if row:
                    if {'west', 'east', 'south', 'north'}.issubset(df.columns):
                        return {
                            'type': 'bbox',
                            'specifier': value,
                            'west': float(row['west']),
                            'east': float(row['east']),
                            'south': float(row['south']),
                            'north': float(row['north'])
                        }

                    if {'lat', 'lon'}.issubset(df.columns):
                        return {
                            'type': 'point',
                            'specifier': value,
                            'lat': float(row['lat']),
                            'lon': float(row['lon'])
                        }

            elif location.suffix == '.nc':
                ds = open_dataset(location, load=settings.LOAD)

                for mask_var in [value, f'm_{value}']:
                    if mask_var in ds.data_vars:
                        return {
                            'type': 'mask',
                            'specifier': value,
                            'mask_ds': ds,
                            'mask_var': mask_var
                        }

            elif location.suffix == '.zip' or location.suffix == '.shp':
                import geopandas
                df = geopandas.read_file(location)
                row = find_row(df, value)
                if row:
                    return {
                        'type': 'shape',
                        'specifier': f'layer-{value}' if value.isdigit() else value,
                        'df': df,
                        'layer': row.name
                    }

    # if region could not be determined, log error and return
    logger.warning(f'could not determine type for region "{value}"')
    return {
        'type': 'unknown',
        'specifier': value
    }


def find_row(df, value):
    if value.isdigit():
        rows = df.loc[df.index == int(value)]
    elif 'specifier' in df.columns:
        rows = df[df['specifier'] == value]
    else:
        rows = pd.DataFrame()

    if not rows.empty:
        return rows.iloc[0].to_dict()


def parse_date(string, start=True):
    try:
        return datetime.strptime(string, "%Y")
    except ValueError:
        try:
            return datetime.strptime(string, "%Y%m%d")
        except ValueError as e:
            raise RuntimeError(f"Unrecognized date format: {string}") from e


def update_path(path, period, region, aggregation, plot=None, start_year=None, end_year=None):
    stem = path.stem

    if aggregation.specifier == 'value':
        region_specifier = region.specifier
    else:
        region_specifier = f'{region.specifier}-{aggregation.specifier}'

    if '_global_' in stem:
        stem = stem.replace('_global_', f'_{region_specifier}_')
    else:
        stem = f'{stem}_{region_specifier}'

    if period.specifier != 'auto':
        stem = re.sub(r'(\d{4}_\d{4}|\d{4})', period.specifier, stem)

    if plot:
        stem += f'_{plot}'

    if start_year:
        stem += f'_{start_year}'

    if end_year:
        stem += f'_{end_year}'

    return path.with_stem(stem)



import logging
import re
from pathlib import Path

from isimip_utils.fetch import fetch_file, load_file
from isimip_utils.files import find_files
from isimip_utils.parameters import (
    apply_placeholders,
    copy_placeholders,
    get_permutations,
    get_placeholders,
)
from isimip_utils.utils import cached_property, get_max_value, get_min_value

from .config import settings
from .utils import init_period, init_region, update_path

logger = logging.getLogger(__name__)


class Dataset:

    def __init__(self, path_template, placeholders):
        self.path = apply_placeholders(path_template, placeholders)
        self.path_template = path_template
        self.placeholders = placeholders

    def  __repr__(self):
        return str(self.path)

    @cached_property
    def full_path(self):
        if settings.DATASETS_PATH:
            return settings.DATASETS_PATH / self.path

    @cached_property
    def files(self):
        if self.full_path:
            glob = sorted(self.full_path.parent.glob(f'{self.full_path.stem}*'))

            return [
                File(file_path, start_year, end_year)
                for file_path, start_year, end_year in find_files(glob)
            ]
        else:
            return []

    @cached_property
    def start_year(self):
        return get_min_value([file.start_year for file in self.files])

    @cached_property
    def end_year(self):
        return get_max_value([file.end_year for file in self.files])

    def exists(self):
        return bool(self.files)

    @classmethod
    def all(cls):
        datasets = []

        if settings.PARAMETERS:
            placeholder_permutations = get_permutations(settings.PARAMETERS)

            for path in settings.PATHS:
                for permutations in placeholder_permutations:
                    placeholders = get_placeholders(settings.PARAMETERS, permutations)
                    datasets.append(cls(path, placeholders))

            return datasets
        else:
            for path in settings.PATHS:
                if re.search(r'\{.*\}', str(path)):
                    raise RuntimeError('Some of the placeholders are missing.')
                datasets.append(cls(path, {}))

        return datasets


class File:

    def __init__(self, path, start_year, end_year):
        self.path = path
        self.start_year = start_year
        self.end_year = end_year

    def  __repr__(self):
        return str(self.path)


class Extraction:

    def __init__(self, dataset, period, region, aggregation):
        self.dataset = dataset
        self.period = period
        self.region = region
        self.aggregation = aggregation

    def  __repr__(self):
        return str(self.path)

    @cached_property
    def path(self):
        path_prefix = update_path(
            self.dataset.path, self.period, self.region, self.aggregation,
            start_year=self.dataset.start_year, end_year=self.dataset.end_year
        )

        if self.dataset.exists():
            # if the dataset exists locally, the prefix was constructed from the actual files, with the correct years
            return path_prefix.with_suffix('.nc')
        elif settings.EXTRACTIONS_PATH:
            # construct a base path for the extraction from the dataset and try to find extraction locally
            full_path_prefix = settings.EXTRACTIONS_PATH / path_prefix
            glob = sorted(full_path_prefix.parent.glob(f'{full_path_prefix.stem}*'))
            for file_path, _, _ in find_files(glob):
                return file_path.relative_to(settings.EXTRACTIONS_PATH)

            # try to find the extraction online
            if settings.EXTRACTIONS_LOCATIONS:
                for location in settings.EXTRACTIONS_LOCATIONS:
                    if isinstance(location, Path):
                        pass
                    else:
                        pattern = path_prefix.stem + r'_(?P<start_year>\d{4})_(?P<end_year>\d{4})?\.nc\d*'
                        index = fetch_file(f'{location}/{path_prefix.parent}')
                        match = re.search(pattern, index)
                        if match:
                            return path_prefix.parent / match.group()

    @cached_property
    def full_path(self):
        if self.path:
            return settings.EXTRACTIONS_PATH / self.path

    def exists(self):
        if self.path:
            return self.full_path.exists()

    def fetch(self):
        if self.path:
            if settings.EXTRACTIONS_LOCATIONS:
                for location in settings.EXTRACTIONS_LOCATIONS:
                    if isinstance(location, Path):
                        result = load_file(location / self.path, self.full_path)
                    else:
                        result = fetch_file(f'{location}/{self.path}', self.full_path)

                    if result:
                        logger.info(f'Extraction "{self.path}" downloaded')
                        return

    @classmethod
    def gather(cls, dataset, periods, regions, aggregations):
        return [
            cls(dataset, period, region, aggregation)
            for period in periods
            for region in regions
            for aggregation in aggregations
        ]


class Figure:

    def __init__(self, path_template, placeholders, period, region, aggregation, plot):
        self.path_template = path_template
        self.placeholders = placeholders
        self.period = period
        self.region = region
        self.aggregation = aggregation
        self.plot = plot

    def  __repr__(self):
        return str(self.path)

    @cached_property
    def path(self):
        if settings.FIGURE_PATH:
            figure_path_template = settings.FIGURE_PATH / Path(self.path_template).name
        else:
            # find the  placeholder which is not in FIGS_PLACEHOLDERS or GRID_PLACEHOLDERS
            figure_path_template = self.path_template
            parts = Path(self.path_template).parts
            name = Path(self.path_template).name
            for i, part in enumerate(parts):
                match = re.match(r"^\{([^}]+)\}$", part)
                if match:
                    if match.group(1) not in settings.FIGURE_PLACEHOLDERS + settings.GRID_PLACEHOLDERS:
                        figure_path_template = Path(*parts[:i]) / name
                        break

        path = update_path(
            apply_placeholders(figure_path_template, self.placeholders),
            self.period, self.region, self.aggregation, self.plot,
            start_year=self.start_year, end_year=self.end_year
        )

        return path.with_suffix(f'.{settings.PLOT_FORMAT}')

    @cached_property
    def full_path(self):
        return settings.PLOTS_PATH / self.path

    @cached_property
    def datasets(self):
        datasets = []
        for grid_permutation in settings.GRID_PERMUTATIONS:
            dataset_placeholders = copy_placeholders(
                self.placeholders,
                get_placeholders(settings.GRID_PARAMETERS, grid_permutation)
            )
            datasets.append(Dataset(self.path_template, dataset_placeholders))
        return datasets

    @cached_property
    def start_year(self):
        return get_min_value([dataset.start_year for dataset in self.datasets])

    @cached_property
    def end_year(self):
        return get_max_value([dataset.end_year for dataset in self.datasets])

    def exists(self):
        return self.full_path.exists()


class Period:

    def __init__(self, value):
        if value == 'auto':
            self.type = self.specifier = value
        else:
            self.__dict__.update(init_period(value))

    def  __repr__(self):
        return self.specifier


class Region:

    def __init__(self, value):
        if value == 'global':
            self.type = self.specifier = value
        else:
            self.__dict__.update(init_region(value))

    def  __repr__(self):
        return self.specifier


class Aggregation:

    def __init__(self, value):
        self.type = self.specifier = value

    def  __repr__(self):
        return self.specifier


class Plot:

    def __init__(self, value):
        self.type = self.specifier = value

        if value == 'map':
            self.columns = {'lat', 'lon'}
        else:
            self.columns = {'time'}

    def  __repr__(self):
        return self.specifier
